using System.Net;
using Common;
using DotNext;
using WebProtocol;

namespace JtagController;

/// <summary>
/// Global Constant Jtag Address
/// </summary>
public static class JtagAddr
{
    /// <summary>
    /// Jtag State Reg
    /// </summary>
    public const UInt32 STATE = 0x10_00_00_00;
    /// <summary>
    /// Jtag Read Reg
    /// </summary>
    public const UInt32 READ_DATA = 0x10_00_00_01;
    /// <summary>
    /// Jtag Write Reg
    /// </summary>
    public const UInt32 WRITE_DATA = 0x10_00_00_02;
    /// <summary>
    /// Jtag Write Command
    /// </summary>
    public const UInt32 WRITE_CMD = 0x10_00_00_03;
}

/// <summary>
/// The Command bits of Jtag
/// </summary>
public static class JtagCmd
{

    /// <summary>
    /// 旁路指令
    /// </summary>
    public const UInt32 JTAG_DR_BYPASS = 0b1111111111;
    /// <summary>
    /// 采样指令
    /// </summary>
    public const UInt32 JTAG_DR_SAMPLE = 0b1010000000;
    /// <summary>
    /// 预装指令
    /// </summary>
    public const UInt32 JTAG_DR_PRELOAD = 0b1010000000;
    /// <summary>
    /// 外测试指令
    /// </summary>
    public const UInt32 JTAG_DR_EXTEST = 0b1010000001;
    /// <summary>
    /// 内测试指令
    /// </summary>
    public const UInt32 JTAG_DR_INTEST = 0b1010000010;
    /// <summary>
    /// 标识指令
    /// </summary>
    public const UInt32 JTAG_DR_IDCODE = 0b1010000011;
    /// <summary>
    /// 高阻指令
    /// </summary>
    public const UInt32 JTAG_DR_HIGHZ = 0b1010000101;
    /// <summary>
    /// 复位指令
    /// </summary>
    public const UInt32 JTAG_DR_JRST = 0b1010001010;
    /// <summary>
    /// 配置指令
    /// </summary>
    public const UInt32 JTAG_DR_CFGI = 0b1010001011;
    /// <summary>
    /// 回读指令
    /// </summary>
    public const UInt32 JTAG_DR_CFGO = 0b1010001100;
    /// <summary>
    /// 唤醒指令
    /// </summary>
    public const UInt32 JTAG_DR_JWAKEUP = 0b1010001101;
    /// <summary>
    /// 读UID指令
    /// </summary>
    public const UInt32 JTAG_DR_READ_UID = 0b0101001100;
    /// <summary>
    /// 读状态寄存器指令
    /// </summary>
    public const UInt32 JTAG_DR_RDSR = 0b0101011001;



    /// <summary>
    /// 设定JTAG默认状态为TEST_LOGIC_RESET态 （JTAG复位）
    /// </summary>
    public const UInt32 CMD_JTAG_CLOSE_TEST = 0b0000;
    /// <summary>
    /// 设定JTAG默认状态为RUN_TEST_IDLE态    （JTAG空闲）
    /// </summary>
    public const UInt32 CMD_JTAG_RUN_TEST = 0b0001;
    /// <summary>
    /// JTAG进入SHIFTIR循环cycle_num次后回到设定的默认状态
    /// </summary>
    public const UInt32 CMD_JTAG_LOAD_IR = 0b0010;
    /// <summary>
    /// JTAG进入SHIFTDR循环cycle_num次后回到设定的默认状态，同时输入shift_in
    /// </summary>
    public const UInt32 CMD_JTAG_LOAD_DR_CAREI = 0b0011;
    /// <summary>
    /// JTAG进入SHIFTDR循环cycle_num次后回到设定的默认状态，同时输出shift_out
    /// </summary>
    public const UInt32 CMD_JTAG_LOAD_DR_CAREO = 0b0100;
    /// <summary>
    /// JTAG进入RUN_TEST_IDLE态循环cycle_num次
    /// </summary>
    public const UInt32 CMD_JTAG_IDLE_DELAY = 0b0101;
}

class Jtag
{
    readonly int timeout;

    readonly int port;
    readonly string address;
    private IPEndPoint ep;
    private UDPServer server;

    public Jtag(string address, int port, UDPServer udpServer, int outTime = 2000)
    {
        this.address = address;
        this.port = port;
        this.ep = new IPEndPoint(IPAddress.Parse(address), port);
        this.server = udpServer;
        this.timeout = outTime;
    }

    public async ValueTask<Result<bool>> RunCommand(uint devAddr, uint cmd, uint exRet)
    {
        var ret = false;
        var opts = new SendAddrPackOptions();

        opts.burstType = BurstType.FixedBurst;
        opts.burstLength = 4;
        opts.commandID = 0;
        opts.address = devAddr;

        // Write Jtag State Register
        opts.isWrite = true;
        ret = await UDPClientPool.SendAddrPackAsync(ep, new SendAddrPackage(opts));
        if (!ret) throw new Exception("Send 1st Address Package Failed!");
        ret = await UDPClientPool.SendDataPackAsync(ep,
                new SendDataPackage(NumberProcessor.NumberToBytes(cmd, 4).Value));
        if (!ret) throw new Exception("Send Data Package Failed!");

        // Read Jtag State Register
        opts.address = JtagAddr.STATE;
        ret = await UDPClientPool.SendAddrPackAsync(ep, new SendAddrPackage(opts));
        if (!ret) throw new Exception("Send 2rd Address Package Failed!");

        // Wait for Ack
        if (!MsgBus.IsRunning)
            throw new Exception("Message Bus not Working!");

        var data = await MsgBus.UDPServer.FindDataAsync(address);
        if (!data.HasValue)
            throw new Exception("Get None after Time out!");

        var recvData = data.Value;
        if (recvData.addr != address || recvData.port != port)
            throw new Exception("Receive Data From Wrong Board!");

        var retPack = RecvDataPackage.FromBytes(recvData.data);
        if (!retPack.IsSuccessful)
            throw new Exception("Not Current RecvDataPackage!", retPack.Error);

        var retPackLen = retPack.Value.Options.data.Length;
        if (retPackLen != 3)
            throw new Exception($"RecvDataPackage BodyData Length not Equal to 3: Total {retPackLen} bytes");

        if (NumberProcessor.BytesToNumber(retPack.Value.Options.data).Value == exRet)
            ret = true;

        return ret;
    }

    public async ValueTask<Result<bool>> ClearAllRegisters()
    {
        return await RunCommand(JtagAddr.STATE, 0xFF_FF_FF_FF, 0x00_00_00_00);
    }

    public async ValueTask<Result<bool>> ClearWriteDataReg()
    {
        return await RunCommand(JtagAddr.STATE, 0x00_00_11_00, 0x00_00_00_00);
    }

    public async ValueTask<Result<bool>> RunTest()
    {
        return await RunCommand(JtagAddr.WRITE_CMD, 0x10_00_00_00, 0x01_00_00_00);
    }

    public async ValueTask<Result<bool>> WriteIDCode()
    {
        return await RunCommand(JtagAddr.WRITE_DATA, 0b1010000011, 0x01_00_00_00);
    }

    public async ValueTask<Result<bool>> LoadIR()
    {
        return await RunCommand(JtagAddr.WRITE_CMD, 0x20_00_00_0A, 0x01_00_00_00);
    }

}
